#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "shape.h"
#include "shape_impl.h"
#include "polygon.h"
#include "polygon_impl.h"
#include "rectangle_impl.h"
#include "coordinates.h"

static void polygon_move(Shape * s, double dx, double dy) {
    Polygon * this = (Polygon *)s;
    for (unsigned int i = 0; i < this->vertices_count; ++i) {
	this->vertices[i].x += dx;
	this->vertices[i].y += dy;
    }
}

static int polygon_print(const Shape * s, FILE * f) {
    const Polygon * this = (const Polygon *)s;

    int res = fprintf(f, "polygon:");
    for (unsigned int i = 0; i < this->vertices_count; ++i) {
	res += fprintf(f, " (%f,%f)", this->vertices[i].x, this->vertices[i].y);
    }
    return res;
}

static void polygon_destroy(Shape * s) {
    Polygon * this = (Polygon *)s;
    free(this->vertices);
    free(this);
}

static void polygon_bounding_rectangle(const Shape * s, Shape * r) {
    const Polygon * this = (const Polygon *)s;
    Rectangle * bbox = (Rectangle *)r;
    Coordinates bottom_left, top_right;
    if (this->vertices_count > 0) {
	bottom_left.x = top_right.x = this->vertices[0].x;
	bottom_left.y = top_right.y = this->vertices[0].y;
    }
    for (unsigned int i = 1; i < this->vertices_count; ++i) {
	if (this->vertices[i].x < bottom_left.x)
	    bottom_left.x = this->vertices[0].x;
	else if  (this->vertices[i].x > top_right.x)
	    top_right.x = this->vertices[0].x;

	if (this->vertices[i].y < bottom_left.y)
	    bottom_left.y = this->vertices[0].y;
	else if  (this->vertices[i].y > top_right.y)
	    top_right.y = this->vertices[0].y;
    }
    bbox->sshape.center.x = (bottom_left.x + top_right.x) / 2;
    bbox->sshape.center.y = (bottom_left.y + top_right.y) / 2;
    bbox->width = top_right.x - bottom_left.x;
    bbox->height = top_right.y - bottom_left.y;
}

static double polygon_area(const Shape * s) {
    const Polygon * this = (const Polygon *)s;
    if (this->vertices_count < 3)
	return 0;
    double a = 0;
    Coordinates * q = this->vertices + (this->vertices_count - 1);
    Coordinates * p = this->vertices;
    for (unsigned int i = 0; i < this->vertices_count; ++i) {
	a += (q->x * p->y) - (p->x * q->y);
	q = p;
	p += 1;
    }
    return a < 0 ? -a/2 : a/2;
}

static struct shape_vtable polygon_vtable = {
    .destroy = polygon_destroy,
    .print = polygon_print,
    .bounding_rectangle = polygon_bounding_rectangle,
    .area = polygon_area,
    .move = polygon_move };

Shape * polygon_new(const Coordinates * begin, const Coordinates * end) {
    Polygon * this = malloc(sizeof(Polygon));
    if (this) {
	shape_init(&(this->shape), &polygon_vtable);

	this->vertices = malloc(sizeof(Coordinates) * (end - begin));
	if (! this->vertices) {
	    free(this);
	    return 0;
	}
	this->vertices_count = end - begin;
	memcpy(this->vertices, begin, sizeof(Coordinates) * (end - begin));
    }
    return &(this->shape);
};

/* Methods 
 */
const Coordinates * polygon_vertices_begin(const Shape * s) {
    const Polygon * this = (const Polygon *)s;
    return this->vertices;
}
    

const Coordinates * polygon_vertices_end(const Shape * s) {
    const Polygon * this = (const Polygon *)s;
    return this->vertices + this->vertices_count;
}

